import pytest
import vishnu
import webapp2


class BaseHandler(webapp2.RequestHandler):

    @property
    def session(self):
        """
        :return: the current vishnu session
        :rtype: vishnu.session.Session
        """
        return vishnu.get_session()


class PrivateHandler(BaseHandler):

    def get(self):

        if self.session.get("user") == "james":
            self.response.status = 200
        else:
            self.response.status = 401


class PublicHandler(BaseHandler):

    def get(self):
        self.response.status = 200


class LoginSaveHandler(BaseHandler):

    def post(self):

        self.session["user"] = "james"
        self.session.save()

        self.response.status = 200


class LoginNoSaveHandler(BaseHandler):

    def post(self):

        self.session["user"] = "james"

        self.response.status = 200


class LogoutHandler(BaseHandler):

    def get(self):

        self.session.terminate()

        self.response.status = 200


@pytest.fixture
def test_app(cookie_name=None, encrypt=False, auto_save=False,
             secure=True, domain=None, path=None,
             use_https=False, backend=None):
    from webtest import TestApp
    from vishnu.middleware import SessionMiddleware

    app = webapp2.WSGIApplication([
        (r'/private', PrivateHandler),
        (r'/public', PublicHandler),
        (r'/login/save', LoginSaveHandler),
        (r'/login/no/save', LoginNoSaveHandler),
        (r'/logout', LogoutHandler)
    ], debug=True)

    encrypt_key = None
    if encrypt:
        encrypt_key = "YTWRsQIU4lYj4HyP33Uh24nrraDFv0R9"

    config = vishnu.session.Config(
        secret="OVc1Mbt79AK5Pmi6sWnJnXZvEPNO3BnI",
        cookie_name=cookie_name,
        encrypt_key=encrypt_key,
        auto_save=auto_save,
        secure=secure,
        backend=backend
    )
    session = SessionMiddleware(app, config)

    extra_environ = {}
    if use_https:
        extra_environ['wsgi.url_scheme'] = 'https'

    return TestApp(app=session, extra_environ=extra_environ)



import pytest


def test_default():
    from vishnu.backend import PyMemcache
    from vishnu.backend.config.memcache import DEFAULT_HOST, DEFAULT_PORT

    config = PyMemcache()
    assert config.host == DEFAULT_HOST
    assert config.port == DEFAULT_PORT


def test_custom_host():
    from vishnu.backend import PyMemcache
    from vishnu.backend.config.memcache import DEFAULT_PORT

    custom_host = "memcache.cloud"
    config = PyMemcache(host=custom_host)
    assert config.host == custom_host
    assert config.port == DEFAULT_PORT


def test_invalid_host():
    from vishnu.backend import PyMemcache

    with pytest.raises(TypeError) as exp:
        PyMemcache(host=23)


def test_custom_port():
    from vishnu.backend import PyMemcache
    from vishnu.backend.config.memcache import DEFAULT_HOST

    custom_port = 11212
    config = PyMemcache(port=custom_port)
    assert config.host == DEFAULT_HOST
    assert config.port == custom_port


def test_invalid_port():
    from vishnu.backend import PyMemcache

    with pytest.raises(TypeError) as exp:
        PyMemcache(port="string")

    with pytest.raises(TypeError) as exp:
        PyMemcache(port=-100)




# -*- coding: utf-8 -*-
import pytest


@pytest.fixture
def cipher():
    from vishnu.cipher import AESCipher
    return AESCipher(key="Qv5p0cmYqkcWl0PyOepbFweKxRY0xAiB")


def test_pad_unpad():
    from vishnu.cipher import AESCipher
    from Crypto.Cipher import AES

    data_to_pad = "secret_raw_data"

    padded_data = AESCipher.pad(data_to_pad)
    assert len(padded_data) % AES.block_size == 0
    unpadded_data = AESCipher.unpad(padded_data)
    assert len(data_to_pad) == len(unpadded_data)
    assert data_to_pad == unpadded_data


def test_encrypt_decrypt(cipher):
    data_to_encrypt = "secret_raw_data"

    encrypted = cipher.encrypt(data_to_encrypt)
    decrypted = cipher.decrypt(encrypted)

    assert data_to_encrypt == decrypted



"""
WSGI Middleware.
"""
from __future__ import absolute_import

from vishnu import _thread_local
from vishnu.session import Session


class SessionMiddleware(object):  # pylint: disable=R0903
    """WSGI middleware for adding support for sessions."""

    def __init__(self, app, config):
        self.app = app
        self.config = config

    def __call__(self, environ, start_response):

        _thread_local.session = Session(environ, self.config)

        def vishnu_start_response(status, headers, exc_info=None):
            """
            Our start_response wrapper so we can insert cookie header
            """
            if _thread_local.session.needs_save and _thread_local.session.auto_save:
                _thread_local.session.save()
            elif _thread_local.session.needs_save and _thread_local.session.started:
                _thread_local.session.save(sync_only=True)

            header = _thread_local.session.header()
            if header:
                headers.append(("Set-Cookie", header))

            return start_response(status, headers, exc_info)

        return self.app(environ, vishnu_start_response)



"""
Vishnu: A session library for the python runtime of Google App Engine.
"""
from __future__ import absolute_import

import threading

__version_info__ = (2, 0, 1)
__version__ = '.'.join(str(v) for v in __version_info__)

_thread_local = threading.local()  # pylint: disable=C0103
_thread_local.session = None


def get_session():
    """Returns the session for the current request"""
    return _thread_local.session



"""
Vishnu session.
"""

from __future__ import absolute_import

from Cookie import SimpleCookie

from datetime import datetime, timedelta
import hashlib
import hmac
import logging
import uuid

from vishnu.cipher import AESCipher
from vishnu.backend.config import Base as BackendConfig

# constant used for specifying this cookie should expire at the end of the session
TIMEOUT_SESSION = "timeout_session"

SECRET_MIN_LEN = 32
ENCRYPT_KEY_MIN_LEN = 32

DEFAULT_COOKIE_NAME = "vishnu"
DEFAULT_PATH = "/"

SIG_LENGTH = 128
SID_LENGTH = 32
EXPIRES_FORMAT = "%a, %d-%b-%Y %H:%M:%S GMT"


class Config(object):

    def __init__(self, secret, cookie_name=None, encrypt_key=None,
                 secure=True, domain=None, path=None, http_only=True,
                 auto_save=False, timeout=None, backend=None):

        self._secret = secret
        if self._secret is None or len(self._secret) < SECRET_MIN_LEN:
            raise ValueError("Secret should be at least %i characters" % SECRET_MIN_LEN)

        if cookie_name is None:
            cookie_name = DEFAULT_COOKIE_NAME

        # todo: check cookie name is a string

        self._cookie_name = cookie_name

        self._encrypt_key = encrypt_key
        if self._encrypt_key is not None and len(self._encrypt_key) < ENCRYPT_KEY_MIN_LEN:
            raise ValueError("Encrypt key should be at least %i characters" % ENCRYPT_KEY_MIN_LEN)

        # todo: check secure is a bool

        self._secure = secure

        # todo: check domain is a string

        self._domain = domain

        if path is None:
            path = DEFAULT_PATH

        # todo: check path is a string

        self._path = path

        # todo: check http_only is a bool

        self._http_only = http_only

        # todo: check auto save is a bool

        self._auto_save = auto_save

        self._timeout = None
        if timeout is not None:
            try:
                self._timeout = int(timeout)
            except ValueError:
                raise TypeError("timeout must be a non-negative integer")

            if self._timeout < 0:
                raise TypeError("timeout must be a non-negative integer")

        if backend is None or not isinstance(backend, BackendConfig):
            raise TypeError("unknown backend configuration received")
        self._backend = backend

    @property
    def secret(self):
        """
        :return: secret used for HMAC signature
        :rtype: string
        """
        return self._secret

    @property
    def cookie_name(self):
        """
        :return: the name for the cookie
        :rtype: string
        """
        return self._cookie_name

    @property
    def encrypt_key(self):
        """
        :return: key to use for encryption
        :rtype: string
        """
        return self._encrypt_key

    @property
    def secure(self):
        """
        :return: whether the cookie can only be transmitted over HTTPS
        :rtype: boolean
        """
        return self._secure

    @property
    def domain(self):
        """
        :return: the domain the cookie is valid for
        :rtype: string
        """
        return self._domain

    @property
    def path(self):
        """
        :return: the path the cookie is valid for
        :rtype: string
        """
        return self._path

    @property
    def http_only(self):
        """
        :return: whether the cookie should only be sent over HTTP/HTTPS
        :rtype: boolean
        """
        return self._http_only

    @property
    def auto_save(self):
        """
        :return: whether this session should auto save
        :rtype: boolean
        """
        return self._auto_save

    @property
    def timeout(self):
        return self._timeout

    @property
    def backend(self):
        """
        :return: config for desired backend
        :rtype: vishnu.backend.config.Base
        """
        return self._backend


class Session(object):
    """The vishnu session object."""

    def __init__(self, environ, config):

        self._environ = environ

        self._send_cookie = False
        self._expire_cookie = False
        self._started = False

        self._sid = Session.generate_sid()
        self._loaded = False
        self._needs_save = False

        # todo: check config is correct class
        self._config = config

        # attempt to load an existing cookie
        self._load_cookie()
        self._backend_client = self._config.backend.client_from_config(self._sid)

        # calculate the expiry date if a timeout exists (must be done after client setup)
        if self._config.timeout:
            self._calculate_expires()




    @classmethod
    def generate_sid(cls):
        """
        :return: generates a unique ID for use by a session
        :rtype: string
        """
        return uuid.uuid4().hex

    @property
    def started(self):
        """
        Has the session been started?
         - True if autosave is on and session has been modified
         - True if autosave is off if session has been saved at least once
         - True is a matching persistent session was found
        """
        return self._started

    @property
    def auto_save(self):
        return self._config.auto_save

    @property
    def needs_save(self):
        """Does this session need to be saved."""
        return self._needs_save

    @property
    def timeout(self):
        """Fetch the current timeout value for this session"""
        return self._config.timeout

    @timeout.setter
    def timeout(self, value):
        """Sets a custom timeout value for this session"""

        if value == TIMEOUT_SESSION:
            self._config.timeout = None
            self._backend_client.expires = None
        else:
            self._config.timeout = value
            self._calculate_expires()

    def _calculate_expires(self):
        """Calculates the session expiry using the timeout"""
        self._backend_client.expires = None

        now = datetime.now()
        self._backend_client.expires = now + timedelta(seconds=self._config.timeout)

    def _load_cookie(self):
        """Loads HTTP Cookie from environ"""

        cookie = SimpleCookie(self._environ.get('HTTP_COOKIE'))
        vishnu_keys = [key for key in cookie.keys() if key == self._config.cookie_name]

        # no session was started yet
        if not vishnu_keys:
            return

        cookie_value = cookie[vishnu_keys[0]].value
        if self._config.encrypt_key:
            cipher = AESCipher(self._config.encrypt_key)
            cookie_value = cipher.decrypt(cookie_value)
        received_sid = Session.decode_sid(self._config.secret, cookie_value)
        if received_sid:
            self._sid = received_sid
        else:
            logging.warn("found cookie with invalid signature")

    def header(self):
        """Generates HTTP header for this cookie."""

        if self._send_cookie:

            cookie_value = Session.encode_sid(self._config.secret, self._sid)
            if self._config.encrypt_key:
                cipher = AESCipher(self._config.encrypt_key)
                cookie_value = cipher.encrypt(cookie_value)

            header = "%s=%s;" % (self._config.cookie_name, cookie_value)

            if self._config.domain:
                header += " Domain=%s;" % self._config.domain

            if self._config.path:
                header += " Path=%s;" % self._config.path

            # expire the cookie
            if self._expire_cookie:
                header += " Expires=Wed, 01-Jan-1970 00:00:00 GMT;"
            # set the cookie expiry
            elif self._backend_client.expires:
                header += " Expires=%s;" % self._backend_client.expires.strftime(EXPIRES_FORMAT)

            if self._config.secure:
                header += " Secure;"
            if self._config.http_only:
                header += " HttpOnly"

            return header
        else:
            return None

    @classmethod
    def encode_sid(cls, secret, sid):
        """Computes the HMAC for the given session id."""
        sig = hmac.new(secret, sid, hashlib.sha512).hexdigest()
        return "%s%s" % (sig, sid)

    @classmethod
    def is_signature_equal(cls, sig_a, sig_b):
        """Compares two signatures using a constant time algorithm to avoid timing attacks."""
        if len(sig_a) != len(sig_b):
            return False

        invalid_chars = 0
        for char_a, char_b in zip(sig_a, sig_b):
            if char_a != char_b:
                invalid_chars += 1
        return invalid_chars == 0

    @classmethod
    def decode_sid(cls, secret, cookie_value):
        """Decodes a cookie value and returns the sid if value or None if invalid."""
        if len(cookie_value) > SIG_LENGTH + SID_LENGTH:
            logging.warn("cookie value is incorrect length")
            return None

        cookie_sig = cookie_value[:SIG_LENGTH]
        cookie_sid = cookie_value[SIG_LENGTH:]
        actual_sig = hmac.new(secret, cookie_sid, hashlib.sha512).hexdigest()

        if not Session.is_signature_equal(cookie_sig, actual_sig):
            return None

        return cookie_sid

    def save(self, sync_only=False):

        # saving a session marks it as started
        self._started = True

        self._backend_client.save(sync_only)

        self._expire_cookie = False
        self._send_cookie = True
        self._needs_save = False

    def terminate(self):
        """Terminates an active session"""
        self._backend_client.clear()
        self._needs_save = False
        self._started = False
        self._expire_cookie = True
        self._send_cookie = True

    def get(self, key):
        """Retrieve a value from the session dictionary"""
        self._started = self._backend_client.load()
        self._needs_save = True

        return self._backend_client.get(key)

    def __setitem__(self, key, value):
        """Set a value in the session dictionary"""
        self._started = self._backend_client.load()
        self._needs_save = True

        # if autosave is on then a session is automatically started when set
        if self._config.auto_save:
            self._started = True

        self._backend_client[key] = value



import pytest


def test_default():
    from vishnu.backend import Redis
    from vishnu.backend.config.redis import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_DB

    config = Redis()
    assert config.host == DEFAULT_HOST
    assert config.port == DEFAULT_PORT
    assert config.db == DEFAULT_DB


def test_custom_host():
    from vishnu.backend import Redis
    from vishnu.backend.config.redis import DEFAULT_PORT, DEFAULT_DB

    custom_host = "memcache.cloud"
    config = Redis(host=custom_host)
    assert config.host == custom_host
    assert config.port == DEFAULT_PORT
    assert config.db == DEFAULT_DB


def test_invalid_host():
    from vishnu.backend import Redis

    with pytest.raises(TypeError) as exp:
        Redis(host=23)


def test_custom_port():
    from vishnu.backend import Redis
    from vishnu.backend.config.redis import DEFAULT_HOST, DEFAULT_DB

    custom_port = 6380
    config = Redis(port=custom_port)
    assert config.host == DEFAULT_HOST
    assert config.port == custom_port
    assert config.db == DEFAULT_DB


def test_invalid_port():
    from vishnu.backend import Redis

    with pytest.raises(TypeError) as exp:
        Redis(port="string")

    with pytest.raises(TypeError) as exp:
        Redis(port=-100)


def test_invalid_db():
    from vishnu.backend import Redis

    with pytest.raises(TypeError) as exp:
        Redis(db=-1)

    with pytest.raises(TypeError) as exp:
        Redis(db="db")


def test_custom_db():
    from vishnu.backend import Redis

    custom_db = 1
    config = Redis(db=custom_db)
    assert config.db == custom_db



import pytest

SECRET = "pFaN6k68lhFkOtNMsc1HfcKvE9GoAJ7Y"
ENCRYPT_KEY = "JBW8bCCKxE2XZ23wUgLJ6T6zFqvPEx1J"


def test_secret():
    from vishnu.session import Config
    from vishnu.backend import Redis

    # short secret
    with pytest.raises(ValueError) as exp:
        short = Config(secret="short")

    # valid secret
    valid = Config(secret=SECRET, backend=Redis())
    assert valid.secret == SECRET


def test_cookie_name():
    from vishnu.backend import Redis
    from vishnu.session import Config
    from vishnu.session import DEFAULT_COOKIE_NAME

    # default
    default = Config(secret=SECRET, backend=Redis())
    assert default.cookie_name == DEFAULT_COOKIE_NAME

    # custom
    custom_cookie_name = "custom_cookie_name"
    custom = Config(secret=SECRET, cookie_name=custom_cookie_name, backend=Redis())
    assert custom.cookie_name == custom_cookie_name


def test_encrypt_key():
    from vishnu.backend import Redis
    from vishnu.session import Config

    # no encrypt key
    default = Config(secret=SECRET, backend=Redis())

    # short encrypt key
    with pytest.raises(ValueError) as exp:
        short = Config(secret=SECRET, encrypt_key="short", backend=Redis())

    # valid encrypt key
    valid = Config(secret=SECRET, encrypt_key=ENCRYPT_KEY, backend=Redis())


def test_secure():
    from vishnu.backend import Redis
    from vishnu.session import Config

    # default
    default = Config(secret=SECRET, backend=Redis())
    assert default.secure == True

    # custom
    custom_true = Config(secret=SECRET, secure=True, backend=Redis())
    assert custom_true.secure == True

    custom_false = Config(secret=SECRET, secure=False, backend=Redis())
    assert custom_false.secure == False


def test_domain():
    from vishnu.backend import Redis
    from vishnu.session import Config

    # default
    default = Config(secret=SECRET, backend=Redis())
    assert default.domain is None

    # custom
    custom_domain = "www.domain.com"
    custom = Config(secret=SECRET, domain=custom_domain, backend=Redis())
    assert custom.domain == custom_domain


def test_path():
    from vishnu.backend import Redis
    from vishnu.session import Config
    from vishnu.session import DEFAULT_PATH

    # default
    default = Config(secret=SECRET, backend=Redis())
    assert default.path == DEFAULT_PATH

    # custom
    custom_path = "/custom"
    custom = Config(secret=SECRET, path=custom_path, backend=Redis())
    assert custom.path == custom_path


def test_http_only():
    from vishnu.backend import Redis
    from vishnu.session import Config

    default = Config(secret=SECRET, backend=Redis())
    assert default.http_only == True

    custom_true = Config(secret=SECRET, http_only=True, backend=Redis())
    assert custom_true.http_only == True

    custom_false = Config(secret=SECRET, http_only=False, backend=Redis())
    assert custom_false.http_only == False


def test_auto_save():
    from vishnu.backend import Redis
    from vishnu.session import Config

    default = Config(secret=SECRET, backend=Redis())
    assert default.auto_save == False

    custom_true = Config(secret=SECRET, auto_save=True, backend=Redis())
    assert custom_true.auto_save == True

    custom_false = Config(secret=SECRET, auto_save=False, backend=Redis())
    assert custom_false.auto_save == False


def test_timeout():
    from vishnu.backend import Redis
    from vishnu.session import Config

    default = Config(secret=SECRET, backend=Redis())
    assert default.timeout is None

    custom_timeout = 1200
    custom = Config(secret=SECRET, timeout=custom_timeout, backend=Redis())
    assert custom.timeout == custom_timeout


def test_backend():
    from vishnu.backend import GoogleAppEngineMemcache
    from vishnu.backend import GoogleAppEngineNDB
    from vishnu.backend import PyMemcache
    from vishnu.backend import PythonMemcached
    from vishnu.backend import Redis
    from vishnu.session import Config

    gae_memcache_config = Config(secret=SECRET, backend=GoogleAppEngineMemcache())

    gae_ndb_config = Config(secret=SECRET, backend=GoogleAppEngineNDB())

    pymemcache_config = Config(secret=SECRET, backend=PyMemcache())

    python_memcached_config = Config(secret=SECRET, backend=PythonMemcached())

    redis_config = Config(secret=SECRET, backend=Redis())

    with pytest.raises(TypeError) as exp:
        unknown_config = Config(secret=SECRET)



"""
Configuration for Google App Engine memcache API.
https://cloud.google.com/appengine/docs/standard/python/memcache/
"""
from vishnu.backend.config import Base


class Config(Base):
    """
    Configuration object for Google App Engine memcache API.

    Should be imported using the shortcut vishnu.backend.GoogleAppEngineMemcache
    """

    def client_from_config(self, sid):
        from vishnu.backend.client.memcache.gae import Client
        return Client(sid)



"""
Terminating and starting a session within same request fails to send cookie
https://github.com/anomaly/vishnu/issues/16
"""
import pytest
import vishnu
import webapp2


class BaseHandler(webapp2.RequestHandler):

    @property
    def session(self):
        """
        :return: the current vishnu session
        :rtype: vishnu.session.Session
        """
        return vishnu.get_session()


class PrivateHandler(BaseHandler):

    def get(self):

        if self.session.get("user") == "james":
            self.response.status = 200
        else:
            self.response.status = 401


class PublicHandler(BaseHandler):

    def get(self):
        self.response.status = 200


class LoginHandler(BaseHandler):

    def post(self):

        self.session["user"] = "james"
        self.session.save()

        self.session.terminate()

        self.session["user"] = "james"
        self.session.save()

        self.response.status = 200


class LogoutHandler(BaseHandler):

    def get(self):

        self.session.terminate()

        self.response.status = 200


@pytest.fixture
def test_app(backend):
    from webtest import TestApp
    from vishnu.middleware import SessionMiddleware
    from vishnu.session import Config

    app = webapp2.WSGIApplication([
        (r'/private', PrivateHandler),
        (r'/public', PublicHandler),
        (r'/login/save', LoginHandler),
        (r'/logout', LogoutHandler)
    ], debug=True)

    config = Config(
        secret="OVc1Mbt79AK5Pmi6sWnJnXZvEPNO3BnI",
        backend=backend
    )
    session = SessionMiddleware(app, config)

    return TestApp(app=session, extra_environ={'wsgi.url_scheme': 'https'})


def test_pymemcache():
    from vishnu.backend import PyMemcache
    app = test_app(backend=PyMemcache())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_python_memcached():
    from vishnu.backend import PythonMemcached
    app = test_app(backend=PythonMemcached())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_redis():
    from vishnu.backend import Redis
    app = test_app(backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401



"""
Client wrapper for Google App Engine NDB
https://cloud.google.com/appengine/docs/standard/python/ndb/
"""
from vishnu.backend.client import Base

from google.appengine.ext import ndb


class VishnuSession(ndb.Model):  # pylint: disable=R0903, W0232
    """NDB model for storing session"""
    expires = ndb.DateTimeProperty(required=False)
    last_accessed = ndb.DateTimeProperty(required=True)
    data = ndb.PickleProperty(required=True, compressed=True)


class Client(Base):
    """
    Client object for Google App Engine NDB
    """

    def __init__(self, sid):
        super(Client, self).__init__(sid)
        self._model = None

    def load(self):
        if self._model is None and not self._loaded:
            self._model = ndb.Key(VishnuSession, self._sid).get()
            if self._model is None:
                return False
            else:
                self._loaded = True

                self.expires = self._model.expires
                self.last_accessed = self._model.last_accessed
                self._data = self._model.data

        return True

    def clear(self):
        self._data = {}
        if self._model:
            self._model.key.delete()
            self._model = None

    def save(self, sync_only=False):
        # try to find an existing session
        self._model = ndb.Key(VishnuSession, self._sid).get()
        if self._model is None:
            self._model = VishnuSession(id=self._sid)

        if sync_only:
            self._model.last_accessed = self.last_accessed
        else:
            self._model.data = self._data
            self._model.last_accessed = self.last_accessed
        if self.expires:
            self._model.expires = self.expires

        self._model.put()



from ..middleware import test_app


def test_manual_save():
    from vishnu.backend import Redis
    app = test_app(use_https=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_auto_save():
    from vishnu.backend import Redis
    app = test_app(use_https=True, auto_save=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/no/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_default_cookie_name():
    from vishnu.backend import Redis
    from vishnu.session import DEFAULT_COOKIE_NAME
    app = test_app(use_https=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200
    assert DEFAULT_COOKIE_NAME in app.cookies

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_custom_cookie_name():
    from vishnu.backend import Redis
    custom_cookie_name = "my-cookie-name"
    app = test_app(use_https=True, cookie_name=custom_cookie_name, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200
    assert custom_cookie_name in app.cookies

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_encrypted():
    from vishnu.backend import Redis
    app = test_app(use_https=True, encrypt=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_insecure_http():
    from vishnu.backend import Redis
    app = test_app(use_https=False, secure=False, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_secure_https():
    from vishnu.backend import Redis
    app = test_app(use_https=True, secure=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private')
    assert private_resp.status_int == 200

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401


def test_secure_http():
    from vishnu.backend import Redis
    app = test_app(use_https=False, secure=True, backend=Redis())

    # check public before login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private before login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # start session (manual save)
    login_resp = app.post("/login/save")
    assert login_resp.status_int == 200

    # check public after login
    public_resp = app.get('/public')
    assert public_resp.status_int == 200

    # check private after login
    private_resp = app.get('/private', status=401)
    assert private_resp.status_int == 401

    # end session
    logout_resp = app.get("/logout")
    assert logout_resp.status_int == 200

    # check public after logout
    public_resp = app.get("/public")
    assert public_resp.status_int == 200

    # check private after logout
    private_resp = app.get("/private", status=401)
    assert private_resp.status_int == 401



"""
Configuration for pymemcache memcache library.
https://pypi.python.org/pypi/pymemcache
"""
from vishnu.backend.config import Base
from vishnu.backend.config.memcache import DEFAULT_HOST, DEFAULT_PORT


class Config(Base):
    """
    Configuration object for pymemcache memcache library.

    Should be imported using the shortcut vishnu.backend.PyMemcache
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT):
        super(Config, self).__init__()

        if not isinstance(host, str):
            raise TypeError("host must be a string")

        self._host = host

        if not isinstance(port, int):
            raise TypeError("port must be a non-negative integer")

        if port < 0:
            raise TypeError("port must be a non-negative integer")

        self._port = port

    @property
    def host(self):
        """
        :return: the host of the memcache instance
        :rtype: string
        """
        return self._host

    @property
    def port(self):
        """
        :return: the port of the memcache instance
        :rtype: integer
        """
        return self._port

    def client_from_config(self, sid):
        from vishnu.backend.client.memcache.pymemcache import Client
        return Client(sid, self._host, self._port)



"""
Client wrapper for python-memcached memcache library
https://pypi.python.org/pypi/python-memcached
"""

from vishnu.backend.client import Base
from vishnu.backend.client import PicklableSession

import memcache
import pickle


class Client(Base):
    """
    Client object for python-memcached memcache library
    """

    def __init__(self, sid, host, port):
        super(Client, self).__init__(sid)
        self._memcache = memcache.Client(["%s:%s" % (host, port)])

        self._record = None

    def load(self):

        if not self._loaded:
            found_in_cache = self._memcache.get(self._sid)

            if found_in_cache is None:
                return False
            else:
                self._record = pickle.loads(found_in_cache)
                self._loaded = True

                self._expires = self._record.expires
                self._last_accessed = self._record.last_accessed
                self._data = self._record.data

        return True

    def clear(self):
        super(Client, self).clear()
        if self._sid:
            self._memcache.delete(self._sid)

    def save(self, sync_only=False):

        # todo: implement sync only

        self._record = PicklableSession(
            self._expires,
            self._last_accessed,
            self._data
        )

        self._memcache.set(self._sid, pickle.dumps(self._record))



import pytest

from vishnu.session import Session
from vishnu.backend import PyMemcache


@pytest.fixture
def memcache_client(sid=None):

    if sid is None:
        sid = Session.generate_sid()

    config = PyMemcache()
    client = config.client_from_config(sid)
    return client


def test_load():
    sid = Session.generate_sid()
    clientA = memcache_client(sid)

    # try to load (not started yet)
    assert clientA.load() == False

    # save
    clientA.save()

    # try to load (should be started)
    assert clientA.load() == True

    # start new client and check already started
    clientB = memcache_client(sid)
    assert clientB.load() == True


def test_clear():
    client = memcache_client()

    # save the session to start it
    client.save()

    # try to load (should be started)
    assert client.load() == True

    # clear
    client.clear()

    # try to load (not started yet)
    assert client.load() == False


def test_save():
    sid = Session.generate_sid()
    clientA = memcache_client(sid)

    assert clientA.load() == False

    clientA.save()

    assert clientA.load() == True

    # save some data to the session
    clientA["key"] = "value"
    clientA.save()

    import time
    time.sleep(1)

    # start another client and check data was loaded
    clientB = memcache_client(sid)

    assert clientB.load() == True
    assert clientB.get("key") == "value"



# memcache
from config.memcache.gae import Config as GoogleAppEngineMemcache
from config.memcache.pymemcache import Config as PyMemcache
from config.memcache.pythonmemcached import Config as PythonMemcached

# redis
from config.redis.redis_py import Config as Redis

# NDB
from config.ndb.gae import Config as GoogleAppEngineNDB



"""
Configuration for Google App Engine NDB.
https://cloud.google.com/appengine/docs/standard/python/ndb/
"""
from vishnu.backend.config import Base


class Config(Base):
    """
    Configuration object for Google App Engine NDB.

    Should be imported using the shortcut vishnu.backend.GoogleAppEngineNDB
    """

    def client_from_config(self, sid):
        from vishnu.backend.client.ndb.gae import Client
        return Client(sid)



# -*- coding: utf-8 -*-
"""
    setup.py
    ~~~~~~~~
    
    A WSGI compliant REST micro-framework.

    :copyright: (c) 2016 Anomaly Software
    :license: Apache 2.0, see LICENSE for more details.
"""

from setuptools import setup, find_packages
from setuptools.command.install import install

from vishnu import __version__

setup(
    name="vishnu",
    version=__version__,
    description='Sessions for the Google App Engine Python runtime',
    url='https://github.com/anomaly/vishnu.git',
    long_description=open("README.rst").read(),
    download_url='https://github.com/anomaly/vishnu/archive/' + __version__ + '.tar.gz',
    license='Apache 2.0',
    author='Anomaly Software',
    author_email='support@anomaly.net.au',
    maintainer='Anomaly Software',
    maintainer_email='support@anomaly.net.au',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Session',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    platforms=['any'],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pycrypto==2.6.1"],
    tests_require=[
        'pytest',
        'redis',
        'pymemcache',
        'python-memcached',
        'webapp2',
        'webtest'
    ],
    setup_requires=['pytest-runner'],
    extras_require={
        'devel': ['tox'],
        'pymemcache': ['pymemcache'],
        'python-memcached': ['python-memcached'],
        'redis': ['redis'],
    }
)



def gae_ndb_delete_expired_sessions(dormant_for=86400, limit=500):
    """Deletes expired sessions
       A session is expired if it expires date is set and has passed or
       if it has not been accessed for a given period of time.
       max_age: seconds since last access to delete sessions, defaults
       to 24 hours.
       limit: amount to delete in one call of the method, the maximum
       and default for this is the NDB fetch limit of 500"""
    from vishnu.backend.client.ndb.gae import VishnuSession
    from google.appengine.ext import ndb
    from datetime import datetime
    from datetime import timedelta

    now = datetime.now()
    last_accessed = now - timedelta(seconds=dormant_for)

    query = VishnuSession.query(ndb.OR(
        ndb.AND(VishnuSession.expires <= now, VishnuSession.expires != None),
        VishnuSession.last_accessed <= last_accessed
    ))
    results = query.fetch(keys_only=True, limit=limit)

    ndb.delete_multi(results)

    return len(results) < limit


"""
Helper classes for encryption.

With a little help from
http://stackoverflow.com/questions/12524994/encrypt-decrypt-using-pycrypto-aes-256
"""

from base64 import b64encode, b64decode
from Crypto.Cipher import AES
from Crypto import Random
import hashlib


class AESCipher(object):
    """
    Helper class for using AES encryption.
    """

    def __init__(self, key):
        self._key = hashlib.sha256(key.encode('utf-8')).digest()

    @classmethod
    def pad(cls, data):
        """
        Pads data to match AES block size
        """
        return data + (AES.block_size - len(data) % AES.block_size) * chr(AES.block_size - len(data) % AES.block_size)

    @classmethod
    def unpad(cls, data):
        """
        Unpads data that has been padded
        """
        return data[:-ord(data[len(data)-1:])]

    def encrypt(self, raw):
        """
        Encrypts raw data using AES and then base64 encodes it.
        :param raw:
        :return:
        """
        padded = AESCipher.pad(raw)
        init_vec = Random.new().read(AES.block_size)
        cipher = AES.new(self._key, AES.MODE_CBC, init_vec)
        return b64encode(init_vec + cipher.encrypt(padded))

    def decrypt(self, encrypted):
        """
        Base64 decodes the data and then decrypts using AES.
        :param encrypted:
        :return:
        """
        decoded = b64decode(encrypted)
        init_vec = decoded[:AES.block_size]
        ciper = AES.new(self._key, AES.MODE_CBC, init_vec)
        return AESCipher.unpad(ciper.decrypt(decoded[AES.block_size:]))



import abc
from datetime import datetime


class PicklableSession(object):

    def __init__(self, expires, last_accessed, data):
        self._expires = expires
        self._last_accessed = last_accessed
        self._data = data

    @property
    def expires(self):
        return self._expires

    @property
    def last_accessed(self):
        return self._last_accessed

    @property
    def data(self):
        return self._data


class Base(object):

    def __init__(self, sid):
        self._sid = sid
        self._loaded = False

        self._expires = None
        self._last_accessed = None

        self._data = {}

    @property
    def expires(self):
        return self._expires

    @expires.setter
    def expires(self, value):
        self._expires = value

    @property
    def last_accessed(self):
        return self._last_accessed

    @last_accessed.setter
    def last_accessed(self, value):
        self._last_accessed = value

    def get(self, key):
        self.last_accessed = datetime.now()
        return self._data.get(key)

    def __setitem__(self, key, value):
        self.last_accessed = datetime.now()
        self._data[key] = value

    @abc.abstractmethod
    def load(self):
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self):
        self._loaded = False

        self._expires = None
        self._last_accessed = None
        self._data = {}

    @abc.abstractmethod
    def save(self, sync_only=False):
        raise NotImplementedError
